/*****************************************************************************/
/*
                                 gZIP.c

Provides ZLIB-enabled GZIP compression (gzip, deflate) for WASD.

Dynamically maps required functions from a ZLIB shareable image.  If the image
is found and all required functions present the ZLIB compression/decompression
of network streams in enabled for suitable requests/responses.  If not it is
just left disabled.

Based on ZLIB v1.2.1 port by Jean-Franois Pironne (jf.pieronne@laposte.net).
Requires this package to be installed and started on the runtime system for
dynamic activation.  The sharable image must be INSTALLed (without any
particular privileges) before it can be activated by the privileged WASD HTTPd
image.  This module will compile on a system that does not have the LIBZ (ZLIB)
header files.  The WASD GZIP.H header contains all of the ZLIB structure and
macro defintions required.  Of course with WASD dynamically activating the LIBZ
sharable image no static linking is required.


RESPONSE CONTENT-ENCODING
-------------------------
The HTTPD$CONFIG directive [GzipResponse] controls whether this feature is
enabled for the gzip content-encoding of suitable response bodies.  This
directive requires at least one parameter, the compression level in the range
1..9.  Smaller values provide faster but poorer compression ratios while larger
values better compression at the cost of more CPU cycles and latency.  This
corresponds to the GZIP utility's -1..-9 CLI switches.  Two optional parameters
could allow ZLIB's 'memLevel' and 'windowBits' to be adjusted by ZLIB
afficiendos (level[,memory,window]).  A small amount of experimentation by this
author indicates minor changes in memory usage and compression ratio by
fiddling with these.

Be aware that GZIP encoding is MEMORY INTENSIVE.  From 132kB to 265kB has been
observed per compressing request (WATCH provides this in a summary line). 
These values apply across a wide range of transfer sizes (from kilobytes to
tens of megabytes).  It also is very CPU INTENSIVE and adds response latency
(on the author's 333MHz system anyway), though that might be well be offset by
significant reductions in transfer time on the Internet or other slower,
non-intranet infrastructures.  Text content compression has been observed from
30% to 10% of the original file size (even down to 1% in the case of the
extremely redundant content of [EXAMPLE]64K.TXT).  VMS executables (for want of
another binary test case) at around 40%.  In other words, GZIP encoding may not
be suitable or efficient for every site or every request!  The other issue is a
"content-encoding: gzip" header necessarily disables the "Content-Length:"
header and this affects the potential for persistent connection maintenance.

Once enabled WASD will GZIP the responses for all suitable contents provided
the client accepts the encoding and the response is not one of the following
(see GzipShouldDeflate()):

  o  less than 512 bytes (no point in the overhead)
  o  already content-encoded script response
  o  script requested that the response not be GZIPed
  o  a compressed image (e.g. GIF, JPEG, etc)
  o  a video stream (presumably already compressed, e.g. MPEG)
  o  a compressed audio stream
  o  an obviously compressed application stream (e.g. GZIP, ZIP, JAR)

Additional control may be exercised with the following path SETings:

  o  response=GZIP=all        matching paths will always have GZIP encoding
                             performed (the above constraints still apply)
  o  response=GZIP=none       matching paths will never have GZIP encoding
  o  response=GZIP=<integer>  responses with content-lengths greater than
                              the specified number of kilobytes will be GZIP
                              content-encoded (if the content-length cannot
                              be determined it will NOT not encoded and the
                              above constraints still apply)


REQUEST CONTENT-ENCODING
------------------------
Decoding of GZIP content-encoded request bodies is enabled using the
HTTPD$CONFIG directive [GzipAccept].  Enabling this using a value 15 (or 1)
results in the server advertising it's acceptance of GZIPed requests using the
"Accept-Encoding: gzip, deflate" response header.  Requests containing bodies
GZIP compressed will have these decoded as they are read from the client and 
before further processing, such as the upload of files into server accessable
file-system space.  THIS DECODING IS OPTIONAL and not the default with DCL and
DECnet script processing.  That is, a request body will be passed to the script
still encoded unless specific mapping directs otherwise.  Decoding by the
server into the original data prior to transfering to the script can be enabled
for all or selected scripts using the following path settings.

  o  script=body=decode       script gets the decoded stream
  o  script=body=NOdecode     script gets the raw, encoded stream (default)

Note that scripts need to be specially aware of both GZIP encoded bodies and
those already decoded by the server.  In the first case the stream must be read
to the specified content-length and then decoded.  In the second case, a
content-length cannot be provided by the server (without unencoding the entire
stream ahead of time it cannot predict the final size).  Where the server is to
decode the request body before transfering it to the script it changes the
CGI variable CONTENT_LENGTH to a single question-mark ("?").  Scripts may use
this to detect the server's intention and then must ignore any
transfer-encoding and/or content-encoding header information and read the
request body until end-of-file is received.

GZIP decoding (decompression) is understandably much less memory and CPU
intensive.  Experimentation indicates it does not contribute significantly to
latency either.


TESTING GZIP
------------
Most modern browsers accept a GZIP deflated response so that's
straight-forward.  Either the document displays correctly or not.  Server
processing of GZIP deflated request bodies is a little more involved.  See the
decription in the BODY.C module.


VERSION HISTORY
---------------
19-DEC-2004  MGD  allow Z_BUF_ERROR with 'ZBufferErrorCount' sanity check
15-NOV-2004  MGD  bugfix; GzipDeflate() (with NET.C)
16-OCT-2004  MGD  initial
*/
/*****************************************************************************/

#ifdef WASD_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#include <stdio.h>
#include <ctype.h>

#include "wasd.h"
#include "gzip.h"

#define WASD_MODULE "GZIP"

/******************/
/* global storage */
/******************/

BOOL  GzipAccept,
      GzipAvailable,
      GzipResponse;

int  GzipDeflateCompLevel,
     GzipDeflateMemLevel,
     GzipDeflateWindowBits,
     GzipFindImageStatus,
     GzipInflateWindowBits;

char  *GzipZlibVersionPtr;

char  *GzipZlibError [] = { "Unknown ZLIB error!!",
                            "ZLIB errno error.",
                            "ZLIB stream error.",
                            "ZLIB data error.",
                            "ZLIB memory error.",
                            "ZLIB buffer error."
                            "ZLIB version error." };

int (*ZlibDeflateFn)(z_stream*, int);
int (*ZlibDeflateInit2Fn)(z_stream*, int, int, int, int, int, char*, int);
int (*ZlibDeflateEndFn)(z_stream*);
int (*ZlibInflateFn)(z_stream*, int);
int (*ZlibInflateInit2Fn)(z_stream*, int, char*, int);
int (*ZlibInflateEndFn)(z_stream*);
char* (*ZlibVersionFn)(void);

/********************/
/* external storage */
/********************/

extern char  ErrorSanityCheck[];

extern ACCOUNTING_STRUCT  *AccountingPtr;
extern CONFIG_STRUCT  Config;
extern WATCH_STRUCT  Watch;

/*****************************************************************************/
/*
Attempt to get the required functions from the ZLIB shareable image.  Any
errors, including the obvious one of image not found, just disables any gzip
encoding.
*/

GzipInit ()

{
   static $DESCRIPTOR (LibzImageDsc, "LIBZ_SHR32");
   static $DESCRIPTOR (DeflateDsc, "deflate");
   static $DESCRIPTOR (DeflateInit2Dsc, "deflateInit2_");
   static $DESCRIPTOR (DeflateEndDsc, "deflateEnd");
   static $DESCRIPTOR (InflateDsc, "inflate");
   static $DESCRIPTOR (InflateInit2Dsc, "inflateInit2_");
   static $DESCRIPTOR (InflateEndDsc, "inflateEnd");
   static $DESCRIPTOR (ZlibVersionDsc, "zlibVersion");

   int  status;

   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (NULL, FI_LI, WATCH_MOD__OTHER, "GzipInit()");

   /* if it's not configured then don't initialize */
   if (!(Config.cfMisc.GzipAcceptWindowBits ||
         Config.cfMisc.GzipResponseCompLevel)) return;

   lib$establish (GzipInitHandler);

   status = lib$find_image_symbol (&LibzImageDsc, &ZlibVersionDsc,
                                   &ZlibVersionFn, 0, 0);
   if (status == RMS$_FNF)
   {
      GzipFindImageStatus = status;
      WriteFaoStdout ("%HTTPD-W-GZIP, sharable image LIBZ_SHR32 not found\n");
      return;
   }

   if (VMSok (status))
      status = lib$find_image_symbol (&LibzImageDsc, &DeflateDsc,
                                      &ZlibDeflateFn, 0, 0);
   if (VMSok (status))
      status = lib$find_image_symbol (&LibzImageDsc, &DeflateInit2Dsc,
                                      &ZlibDeflateInit2Fn, 0, 0);
   if (VMSok (status))
      status = lib$find_image_symbol (&LibzImageDsc, &DeflateEndDsc,
                                      &ZlibDeflateEndFn, 0, 0);
   if (VMSok (status))
      status = lib$find_image_symbol (&LibzImageDsc, &InflateDsc,
                                      &ZlibInflateFn, 0, 0);
   if (VMSok (status))
      status = lib$find_image_symbol (&LibzImageDsc, &InflateInit2Dsc,
                                      &ZlibInflateInit2Fn, 0, 0);
   if (VMSok (status))
      status = lib$find_image_symbol (&LibzImageDsc, &InflateEndDsc,
                                      &ZlibInflateEndFn, 0, 0);
   lib$revert ();

   GzipFindImageStatus = status;
   if (VMSok (status))
   {
      GzipAvailable = true;
      if (Config.cfMisc.GzipAcceptWindowBits) GzipAccept = true;
      if (Config.cfMisc.GzipResponseCompLevel) GzipResponse = true;

      GzipInflateWindowBits = Config.cfMisc.GzipResponseWindowBits;
      if (GzipInflateWindowBits < 8)
         GzipInflateWindowBits = 15;  /* assume the default */
      else
      if (GzipInflateWindowBits > 15)
         GzipInflateWindowBits = 15;

      GzipDeflateCompLevel = Config.cfMisc.GzipResponseCompLevel;
      if (GzipDeflateCompLevel > 9) GzipDeflateCompLevel = 9;

      GzipDeflateMemLevel = Config.cfMisc.GzipResponseMemLevel;
      /* if not specified default to ZLIB default */
      if (!GzipDeflateMemLevel) GzipDeflateMemLevel = 8;
      if (GzipDeflateMemLevel > 9) GzipDeflateMemLevel = 9;

      GzipDeflateWindowBits = Config.cfMisc.GzipResponseWindowBits;
      /* if not specified default to ZLIB default */
      if (!GzipDeflateWindowBits) GzipDeflateWindowBits = 15;
      if (GzipDeflateWindowBits < 8)
         GzipDeflateWindowBits = 15;  /* assume the default */
      else
      if (GzipDeflateWindowBits > 15)
         GzipDeflateWindowBits = 15;

      GzipZlibVersionPtr = ZlibVersionFn();
      WriteFaoStdout ("%HTTPD-I-GZIP, using ZLIB !AZ\n", GzipZlibVersionPtr);
   }
   else
      WriteFaoStdout ("%HTTPD-W-GZIP, fail ZLIB !&S\r\n-!-!&M\n", status);
}

/*****************************************************************************/
/*
Just continue, to report an error if the image couldn't be activated or the
required symbol not found.
*/

int GzipInitHandler ()

{
   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (NULL, FI_LI, WATCH_MOD__OTHER, "GzipInitHandler()");

   return (SS$_CONTINUE);
}

/*****************************************************************************/
/*
Initialize the ZLIB structure.  Allocate memory from the ZLIB output buffer.
Returns true and set the pointer to the ZLIB structure if successful.  Returns
false and leaves the pointer NULL if not.  Reports any error via the usual WASD
mechanism.
*/

BOOL GzipDeflateBegin
(
REQUEST_STRUCT *rqptr,
GZIP_COMPRESS *gzptr,
int BufferSize
)
{
   int  zstatus;
   char  *cptr;
   z_stream  *zsptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (rqptr, FI_LI, WATCH_MOD__OTHER, "GzipDeflateBegin()");

   gzptr->DeflateStartStream = true;
   gzptr->DeflateEndOfStream = false;
   gzptr->DeflateMemoryAllocated = rqptr->GzipMemoryAllocated = 0;

   zsptr = (z_stream*)VmGetHeap (rqptr, sizeof(z_stream));

   zsptr->opaque = rqptr;
   zsptr->zalloc = &GzipAlloc;
   zsptr->zfree = &GzipFree;

   /* plus 16 is the magic number to get a gzip header and trailer!! */
   zstatus = ZlibDeflateInit2Fn (zsptr,
                                 GzipDeflateCompLevel,
                                 Z_DEFLATED,
                                 GzipDeflateWindowBits + 16,
                                 GzipDeflateMemLevel,
                                 Z_DEFAULT_STRATEGY,
                                 ZLIB_VERSION,
                                 sizeof(z_stream));

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (rqptr, FI_LI, WATCH_MOD__OTHER,
                 "ZlibDeflateInit2Fn(!UL,!UL+16,!UL) !SL !UL !&Z",
                 GzipDeflateCompLevel, GzipDeflateWindowBits,
                 GzipDeflateMemLevel, zstatus, gzptr->DeflateMemoryAllocated,
                 zsptr->msg);

   if (zstatus == Z_OK)
   {
      gzptr->DeflateZstreamPtr = zsptr;
      gzptr->DeflateBufferSize = BufferSize;
      gzptr->DeflateBufferPtr = VmGetHeap (rqptr, BufferSize);
      return (true);
   }

   if (!(cptr = zsptr->msg))
   {
      zstatus = -zstatus;
      if (zstatus < 0 || zstatus > 5) zstatus = 0;
      cptr = GzipZlibError[zstatus];
   }
   rqptr->rqResponse.HttpStatus = 500;
   ErrorGeneral (rqptr, cptr, FI_LI);
   return (false);
} 

/*****************************************************************************/
/*
Deallocate ZLIB structures.
Returns true if there was no last-minute problem reported by ZLIB, false and
reports the error if there was.
*/

BOOL GzipDeflateEnd
(
REQUEST_STRUCT *rqptr,
GZIP_COMPRESS *gzptr
)
{
   int  zstatus;
   char  *cptr;
   z_stream  *zsptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
   {
      if (gzptr->DeflateZstreamPtr)
      {
         int  DeflatePercent;
         gzptr->DeflateMemoryAllocated = rqptr->GzipMemoryAllocated;
         if (gzptr->DeflateBytesIn)
            DeflatePercent =  gzptr->DeflateBytesOut * 100 /
                              gzptr->DeflateBytesIn;
         else
            DeflatePercent = 0;
         WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                    "DEFLATE !UL->!UL bytes, !UL% (!ULkB)",
                    gzptr->DeflateBytesIn, gzptr->DeflateBytesOut,
                    DeflatePercent, gzptr->DeflateMemoryAllocated >> 10);
      }
   }

   gzptr->DeflateEndOfStream = true;

   InstanceMutexLock (INSTANCE_MUTEX_HTTPD);
   AccountingPtr->GzipDeflateCount++;
   ADD_LONG_QUAD (gzptr->DeflateBytesIn, AccountingPtr->GzipDeflateBytesIn);
   ADD_LONG_QUAD (gzptr->DeflateBytesOut, AccountingPtr->GzipDeflateBytesOut);
   InstanceMutexUnLock (INSTANCE_MUTEX_HTTPD);

   if (zsptr = gzptr->DeflateZstreamPtr)
   {
      zstatus = ZlibDeflateEndFn (zsptr);
      /* this as NULL indicates that deflate is no longer in use */
      gzptr->DeflateZstreamPtr = NULL;
   }
   else
      zstatus = Z_OK;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (rqptr, FI_LI, WATCH_MOD__OTHER,
                 "ZlibDeflateEnd() !SL", zstatus);

   if (zstatus == Z_OK) return (true);

   zstatus = -zstatus;
   if (zstatus < 0 || zstatus > 5) zstatus = 0;
   cptr = GzipZlibError[zstatus];

   rqptr->rqResponse.HttpStatus = 500;
   ErrorGeneral (rqptr, cptr, FI_LI);
   return (false);
} 

/*****************************************************************************/
/*
Data to be written to the network is intercepted and passed to this function. 
It uses the ZLIB deflate function to compress the data into the WASD buffer
allocated during initialization.

The '*InDataPtrPtr' and '*InDataLengthPtr' point to the input buffer.  If
'*InDataPtrPtr' is NULL then it's end-of-stream.  When input space is available
and the input buffer consumed '*InDataLengthPtr' is modified to be zero.  This
detected by the calling routine to indicate more input data (actually server
output data, but input to GZIP) should be obtained.

The '*OutDataPtrPtr' and '*DataLengthPtr' are then modified to reflect any
compressed data available for writing (zero indicates none and more should be
fed to the input).

Return true to indicate success, false if the write (and request) should be
terminated due to error (which is also reported by the usual WASD mechanism).
*/

BOOL GzipDeflate
(
REQUEST_STRUCT *rqptr,
GZIP_COMPRESS *gzptr,
char **InDataPtrPtr,
int *InDataLengthPtr,
char **OutDataPtrPtr,
int *OutDataLengthPtr
)
{
   int  zflush, zstatus,
        InDataLength,
        OriginalAvailIn;
   char  *cptr,
         *InDataPtr;
   z_stream  *zsptr;

   /*********/
   /* begin */
   /*********/

   zsptr = gzptr->DeflateZstreamPtr;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (rqptr, FI_LI, WATCH_MOD__OTHER,
                 "GzipDeflate() !&B !UL !UL",
                 !*InDataPtrPtr, zsptr->avail_in, zsptr->avail_out);

   /* a NULL pointer indicates end-of-stream */
   if (*InDataPtrPtr)
      zflush = Z_NO_FLUSH;
   else
      zflush = Z_FINISH;

   *OutDataLengthPtr = 0;
   *OutDataPtrPtr = NULL;

   if (*InDataLengthPtr || zsptr->total_in)
   {
      /* if we have data to process or have previously processed some */
      if (!zsptr->avail_in)
      {
         /* requires more input */
         zsptr->next_in = InDataPtr = *InDataPtrPtr;
         zsptr->avail_in = InDataLength = *InDataLengthPtr;
         *InDataLengthPtr = 0;
      }
      if (!zsptr->avail_out)
      {
         /* output buffer needs (re)setting */
         zsptr->next_out = gzptr->DeflateBufferPtr;
         zsptr->avail_out = gzptr->DeflateBufferSize;
      }

      zstatus = ZlibDeflateFn (zsptr, zflush);

      gzptr->DeflateBytesIn = zsptr->total_in;
      gzptr->DeflateBytesOut = zsptr->total_out;

      if (zstatus == Z_OK && !zsptr->avail_out)
         gzptr->DeflateBufferCount = gzptr->DeflateBufferSize;
      else
      if (zstatus == Z_STREAM_END)
         gzptr->DeflateBufferCount = gzptr->DeflateBufferSize -
                                     zsptr->avail_out;
      else
         gzptr->DeflateBufferCount = 0;

      *OutDataPtrPtr = gzptr->DeflateBufferPtr;
      *OutDataLengthPtr = gzptr->DeflateBufferCount;

      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD__OTHER))
         WatchThis (rqptr, FI_LI, WATCH_MOD__OTHER,
                    "!SL !UL/!UL !&X:!UL !&X:!UL !UL->!UL !&Z",
                    zstatus, zsptr->avail_in, zsptr->avail_out, 
                    InDataPtr, InDataLength,
                    *OutDataPtrPtr, *OutDataLengthPtr,
                    gzptr->DeflateBytesIn, gzptr->DeflateBytesOut, zsptr->msg);
   }
   else
   {
      /* no data has ever been processed, just clean up the ZLIB stuff */
      zstatus = Z_STREAM_END;
   }

   if (zstatus == Z_OK)
   {
      gzptr->ZBufferErrorCount = 0;
      return (true);
   }

   if (zstatus == Z_STREAM_END)
      if (GzipDeflateEnd (rqptr, gzptr))
         return (true);

   if (zstatus == Z_BUF_ERROR)
   {
      /* according to the ZLIB doc we're allowed at least one of these */
      if (gzptr->ZBufferErrorCount++ < 3) return (true);
      ErrorNoticed (rqptr, SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
   }

   if (!(cptr = zsptr->msg))
   {
      zstatus = -zstatus;
      if (zstatus < 0 || zstatus > 5) zstatus = 0;
      cptr = GzipZlibError[zstatus];
   }
   rqptr->rqResponse.HttpStatus = 500;
   ErrorGeneral (rqptr, cptr, FI_LI);

   GzipDeflateEnd (rqptr, gzptr);

   return (false);
} 

/*****************************************************************************/
/*
Can and should the content body be deflated?
Return true to indicate deflate, false to leave as-is.
*/

BOOL GzipShouldDeflate
(
REQUEST_STRUCT *rqptr,
char *ContentTypePtr,
int ContentLength
)
{
   BOOL  EncodeGzip;
   char  *cptr;
   char  ResponseKbytes [32];

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (rqptr, FI_LI, WATCH_MOD__OTHER,
                 "GzipShouldDeflate() !&B !&Z !SL",
                 GzipAvailable, ContentTypePtr, ContentLength);

   if (!GzipAvailable) return (false);

   if (!ContentTypePtr) ContentTypePtr = "";
   EncodeGzip = true;

   if (!rqptr->rqHeader.AcceptEncodingGzip)
   {
      EncodeGzip = false;
      cptr = "client does not accept";
   }
   else
   if (rqptr->rqCgi.ContentEncoding)
   {
      EncodeGzip = false;
      cptr = "script response already encoded";
   }
   else
   if (rqptr->rqCgi.ContentEncodingGzip == CGI_OUTPUT_NO_GZIP)
   {
      EncodeGzip = false;
      cptr = "script requested response NOT to be GZIPed";
   }
   else
   if (rqptr->rqPathSet.ResponseGzip == MAPURL_RESPONSE_GZIP_NONE &&
       rqptr->rqCgi.ContentEncodingGzip != CGI_OUTPUT_GZIP)
   {
      /* note that the script requesting GZIP overrides the mapping rule */
      EncodeGzip = false;
      cptr = "response=GZIP=none";
   }
   else
   if (!rqptr->rqPathSet.ResponseGzip &&
       ContentLength >= 0 &&
       ContentLength < GZIP_MIN_CONTENT_LENGTH)
   {
      /* when fairly small avoid the compression overhead */
      EncodeGzip = false;
      cptr = "minimal content-length";
   }
   else
   if (!memcmp (ContentTypePtr, "image/", 6) &&
       memcmp (ContentTypePtr+6, "bmp", 3) &&
       memcmp (ContentTypePtr+6, "tiff", 4) &&
       memcmp (ContentTypePtr+6, "x-bitmap", 8) &&
       memcmp (ContentTypePtr+6, "x-pixmap", 8))
   {
      /* most image content is already compressed (except the above) */
      EncodeGzip = false;
      cptr = "compressed image";
   }
   else
   if (!memcmp (ContentTypePtr, "audio/", 6) &&
       memcmp (ContentTypePtr+6, "x-wav", 5))
   {
      /* generally audio is already compressed (except WAVs AFAIK) */
      EncodeGzip = false;
      cptr = "compressed audio";
   }
   else
   if (!memcmp (ContentTypePtr, "video/", 6))
   {
      /* video is (hopefully!) already compressed */
      EncodeGzip = false;
      cptr = "compressed video";
   }
   else
   if (!memcmp (ContentTypePtr, "application/", 12) &&
       (strstr (ContentTypePtr+12, "zip") ||
        strstr (ContentTypePtr+12, "gtar") ||
        strstr (ContentTypePtr+12, "archive") ||
        strstr (ContentTypePtr+12, "stuffit")))
   {
      /* compressed application stream */
      EncodeGzip = false;
      cptr = "compressed archive";
   }
   else
   if (!memcmp (ContentTypePtr, "multipart/", 10) &&
       strstr (ContentTypePtr+10, "zip"))
   {
      /* same, same */
      EncodeGzip = false;
      cptr = "compressed archive";
   }
   else
   if (rqptr->rqPathSet.ResponseGzip &&
       rqptr->rqPathSet.ResponseGzip != MAPURL_RESPONSE_GZIP_ALL &&
       rqptr->rqPathSet.ResponseGzip != MAPURL_RESPONSE_GZIP_NONE &&
       (ContentLength < 0 ||
        rqptr->rqPathSet.ResponseGzip >= ContentLength >> 10))
   {
      EncodeGzip = false;
      if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
      {
         if (ContentLength >= 0)
            WriteFao (cptr = ResponseKbytes, sizeof(ResponseKbytes), NULL,
                      "response=GZIP=!UL (!ULkB)",
                      rqptr->rqPathSet.ResponseGzip, ContentLength >> 10);
         else
            WriteFao (cptr = ResponseKbytes, sizeof(ResponseKbytes), NULL,
                      "response=GZIP=!UL (?kB)", rqptr->rqPathSet.ResponseGzip);
      }
   }
   else
   if (rqptr->rqResponse.NoGzip)
   {
      EncodeGzip = false;
      cptr = "response NOT to be GZIPed";
   }

   rqptr->rqResponse.ContentEncodingGzip = EncodeGzip;

   if (!EncodeGzip)
      if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
         WatchThis (rqptr, FI_LI, WATCH_RESPONSE, "DEFLATE no, !AZ", cptr);

   return (EncodeGzip);
} 

/*****************************************************************************/
/*
Initialize the ZLIB structure.  Allocate memory from the ZLIB output buffer.
Returns true and set the pointer to the ZLIB structure if successful.  Returns
false and leaves the pointer NULL if not.  Reports any error via the usual WASD
mechanism.
*/

BOOL GzipInflateBegin
(
REQUEST_STRUCT *rqptr,
GZIP_COMPRESS *gzptr,
int BufferSize
)
{
   int  zstatus;
   char  *cptr;
   z_stream  *zsptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (rqptr, FI_LI, WATCH_MOD__OTHER, "GzipInflateBegin()");

   gzptr->InflateStartStream = true;
   gzptr->InflateEndOfStream = false;
   gzptr->InflateMemoryAllocated = rqptr->GzipMemoryAllocated = 0;

   zsptr = (z_stream*)VmGetHeap (rqptr, sizeof(z_stream));

   zsptr->opaque = rqptr;
   zsptr->zalloc = &GzipAlloc;
   zsptr->zfree = &GzipFree;

   /* plus 16 is the magic number to remove a gzip header and trailer!! */
   zstatus = ZlibInflateInit2Fn (zsptr,
                                 GzipInflateWindowBits + 16,
                                 ZLIB_VERSION,
                                 sizeof(z_stream));

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (rqptr, FI_LI, WATCH_MOD__OTHER,
                 "ZlibInflateInit2Fn(!UL+16) !SL !UL !&Z",
                 GzipDeflateWindowBits, zstatus,
                 gzptr->InflateMemoryAllocated, zsptr->msg);

   if (zstatus == Z_OK)
   {
      gzptr->InflateZstreamPtr = zsptr;
      gzptr->InflateBufferSize = BufferSize;
      gzptr->InflateBufferPtr = VmGetHeap (rqptr, BufferSize);
      return (true);
   }

   if (!(cptr = zsptr->msg))
   {
      zstatus = -zstatus;
      if (zstatus < 0 || zstatus > 5) zstatus = 0;
      cptr = GzipZlibError[zstatus];
   }
   rqptr->rqResponse.HttpStatus = 500;
   ErrorGeneral (rqptr, cptr, FI_LI);
   return (false);
} 

/*****************************************************************************/
/*
Deallocate ZLIB structures.
Returns true if there was no last-minute problem reported by ZLIB, false and
reports the error if there was.
*/

BOOL GzipInflateEnd
(
REQUEST_STRUCT *rqptr,
GZIP_COMPRESS *gzptr
)
{
   int  zstatus;
   char  *cptr;
   z_stream  *zsptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_RESPONSE))
   {
      if (gzptr->InflateZstreamPtr)
      {
         int  InflatePercent;
         gzptr->InflateMemoryAllocated = rqptr->GzipMemoryAllocated;
         if (gzptr->InflateBytesIn)
            InflatePercent =  gzptr->InflateBytesIn * 100 /
                              gzptr->InflateBytesOut;
         else
            InflatePercent = 0;
         WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                    "INFLATE !UL->!UL bytes, !UL% (!ULkB)",
                    gzptr->InflateBytesIn, gzptr->InflateBytesOut,
                    InflatePercent, gzptr->InflateMemoryAllocated >> 10);
      }
   }

   gzptr->InflateEndOfStream = true;

   InstanceMutexLock (INSTANCE_MUTEX_HTTPD);
   AccountingPtr->GzipInflateCount++;
   ADD_LONG_QUAD (gzptr->InflateBytesIn, AccountingPtr->GzipInflateBytesIn);
   ADD_LONG_QUAD (gzptr->InflateBytesOut, AccountingPtr->GzipInflateBytesOut);
   InstanceMutexUnLock (INSTANCE_MUTEX_HTTPD);

   if (zsptr = gzptr->InflateZstreamPtr)
   {
      zstatus = ZlibInflateEndFn (zsptr);
      /* this as NULL indicates that inflate is no longer in use */
      gzptr->InflateZstreamPtr = NULL;
   }
   else
      zstatus = Z_OK;

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (rqptr, FI_LI, WATCH_MOD__OTHER,
                 "ZlibInflateEnd() !SL", zstatus);

   if (zstatus == Z_OK)
   {
      gzptr->ZBufferErrorCount = 0;
      return (true);
   }

   if (zstatus == Z_BUF_ERROR)
   {
      /* according to the ZLIB doc we're allowed at least one of these */
      if (gzptr->ZBufferErrorCount++ < 3) return (true);
      ErrorNoticed (rqptr, SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
   }

   zstatus = -zstatus;
   if (zstatus < 0 || zstatus > 5) zstatus = 0;
   cptr = GzipZlibError[zstatus];

   rqptr->rqResponse.HttpStatus = 500;
   ErrorGeneral (rqptr, cptr, FI_LI);
   return (false);
} 

/*****************************************************************************/
/*
Data read from the network is intercepted and passed to this function. 
It uses the ZLIB inflate function to decompress the data into the WASD buffer
allocated during initialization.  The 'DataPtr' and 'DataLength' are then
modified to reflect the decompressed data and it can then be further processed.
Return true to indicate success, false if the write (and request) should be
terminated due to error (which is also reported by the usual WASD mechanism).
*/

BOOL GzipInflate
(
REQUEST_STRUCT *rqptr,
GZIP_COMPRESS *gzptr,
char **DataPtrPtr,
int *DataLengthPtr
)
{
   int  zstatus;
   char  *cptr;
   z_stream  *zsptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (rqptr, FI_LI, WATCH_MOD__OTHER, "GzipInflate() !&X !UL",
                 *DataPtrPtr, *DataLengthPtr);

   zsptr = gzptr->InflateZstreamPtr;

   if (!zsptr->avail_in)
   {
      zsptr->next_in = *DataPtrPtr;
      zsptr->avail_in = *DataLengthPtr;
   }

   if (zsptr->avail_in || zsptr->total_in)
   {
      /* if we have data to process or have previously processed some */
      if (!zsptr->avail_out)
      {
         /* output buffer needs (re)setting */
         zsptr->next_out = gzptr->InflateBufferPtr;
         zsptr->avail_out = gzptr->InflateBufferSize;
      }

      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD__OTHER))
         WatchThis (rqptr, FI_LI, WATCH_MOD__OTHER, "!UL(!UL)/!UL",
                    zsptr->avail_in,  gzptr->InflateAvailIn, zsptr->avail_out);

      zstatus = ZlibInflateFn (zsptr, Z_NO_FLUSH);

      gzptr->InflateBytesIn = zsptr->total_in;
      gzptr->InflateBytesOut = zsptr->total_out;

      if (zstatus == Z_OK && !zsptr->avail_out)
         gzptr->InflateBufferCount = gzptr->InflateBufferSize;
      else
      if (zstatus == Z_STREAM_END)
         gzptr->InflateBufferCount = gzptr->InflateBufferSize -
                                     zsptr->avail_out;
      else
         gzptr->InflateBufferCount = 0;

      if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD__OTHER))
         WatchThis (rqptr, FI_LI, WATCH_MOD__OTHER,
                    "!SL !UL/!UL !UL !UL->!UL !&Z",
                    zstatus, zsptr->avail_in, zsptr->avail_out, 
                    gzptr->InflateBufferCount, gzptr->InflateBytesIn,
                    gzptr->InflateBytesOut, zsptr->msg);
   }
   else
   {
      /* no data has ever been processed, just clean up the ZLIB stuff */
      zstatus = Z_STREAM_END;
      gzptr->InflateBufferCount = 0;
   }

   if (zstatus == Z_OK ||
       zstatus == Z_STREAM_END)
   {
      if (zstatus == Z_STREAM_END)
         if (!GzipInflateEnd (rqptr, gzptr))
            return (false);

      gzptr->InflateAvailIn = zsptr->avail_in;
      *DataPtrPtr = gzptr->InflateBufferPtr;
      *DataLengthPtr = gzptr->InflateBufferCount;

      return (true);
   }

   if (!(cptr = zsptr->msg))
   {
      zstatus = -zstatus;
      if (zstatus < 0 || zstatus > 5) zstatus = 0;
      cptr = GzipZlibError[zstatus];
   }
   rqptr->rqResponse.HttpStatus = 500;
   ErrorGeneral (rqptr, cptr, FI_LI);

   GzipInflateEnd (rqptr, gzptr);

   return (false);
} 

/*****************************************************************************/
/*
Allocate request heap memory on behalf of ZLIB demands.
Shouldn't leak any memory this way (not that ZLIB is likely to).
*/

char* GzipAlloc
(
REQUEST_STRUCT *rqptr,
int items,
int size
)
{
   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (NULL, FI_LI, WATCH_MOD__OTHER,
                 "GzipAlloc() !UL !UL", items, size);

   rqptr->GzipMemoryAllocated += size*items; 
   return (VmGetHeap (rqptr, size*items));
} 

/*****************************************************************************/
/*
Free GzipAlloc()ed request heap memory.
*/

GzipFree
(
REQUEST_STRUCT *rqptr,
char *mptr
)
{
   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD__OTHER))
      WatchThis (NULL, FI_LI, WATCH_MOD__OTHER, "GzipFree()");

   VmFreeFromHeap (rqptr, mptr, FI_LI);
} 

/*****************************************************************************/

