/*****************************************************************************/
/*
                                 ProxyNet.c

*************
** CAUTION **
*************

THIS MODULE IS TASK-ORIENTED, NOT REQUEST-ORIENTED.

That is, most of the functions take a pointer to proxy task rather than a
pointer to request as do other modules. The reason is simple. Many of these
functions are designed for use independently of a request.

This module provides the essential networking functionality for proxy
processing.  It also maintains the pool of persistent proxy->origin server
connections.


VERSION HISTORY
---------------
20-SEP-2004  MGD  extracted functionality from PROXY.C
*/
/*****************************************************************************/

#ifdef WASD_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <dvidef.h>
#include <iodef.h>
#include <psldef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application-related header files */
#include "wasd.h"

#define WASD_MODULE "PROXYNET"

/******************/
/* global storage */
/******************/

int  ProxyNetConnectCount,
     ProxyNetConnectCountMax,
     ProxyNetConnectFreeCount;

LIST_HEAD  ProxyNetConnectList,
           ProxyNetConnectFreeList;

/********************/
/* external storage */
/********************/

extern int  EfnWait,
            EfnNoWait,
            HttpdTickSecond,
            OptionEnabled,
            ProxyConnectPersistMax,
            ProxyConnectPersistSeconds;

extern int  ToLowerCase[],
            ToUpperCase[];

extern unsigned long  HttpdBinTime[];

extern char  ErrorSanityCheck[],
             SoftwareID[],
             Utility[];

extern struct dsc$descriptor TcpIpDeviceDsc;

extern ACCOUNTING_STRUCT  *AccountingPtr;
extern IPADDRESS  TcpIpEmptyAddress;
extern PROXY_ACCOUNTING_STRUCT  *ProxyAccountingPtr;
extern MSG_STRUCT  Msgs;
extern SERVICE_STRUCT  *ServiceListHead;
extern TCP_SOCKET_ITEM  TcpIpSocket4,
                        TcpIpSocket6;
extern VMS_ITEM_LIST2  ReuseAddress,
                       ReuseAddressSocketOption,
                       TcpIpFullDuplexCloseOption;
extern WATCH_STRUCT  Watch;

/****************************************************************************/
/*
*/

ProxyNetInit ()

{
   /*********/
   /* begin */
   /*********/

   if (WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (NULL, FI_LI, WATCH_MOD_PROXY, "ProxyNetInit()");

   if (ProxyConnectPersistMax < 0 || ProxyConnectPersistMax > 1000)
      ProxyConnectPersistMax = 0;

   if (ProxyConnectPersistSeconds < 0 || ProxyConnectPersistSeconds > 3600)
      ProxyConnectPersistSeconds = 0;

   ProxyNetConnectSupervisor (0);
}


/****************************************************************************/
/*
No need to resolve if this proxy server chains to another.  Otherwise, allow
TcpIpNameToAddress() to asynchronously resolve the name or numeric address into
the IP address and then call ProxyNetHostConnect() when ready.
*/

ProxyNetResolveHost (PROXY_TASK *tkptr)

{
   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyNetResolveHost() !&Z", tkptr->RequestHostName);

   if (IPADDRESS_IS_RESET (&tkptr->ChainIpAddress))
   {
      /************************************/
      /* resolve host using TCP/IP lookup */
      /************************************/

      tkptr->ConnectPort = tkptr->RequestPort;
      tkptr->ConnectHostPortPtr = tkptr->RequestHostPort;

      TcpIpNameToAddress (&tkptr->HostLookup,
                          tkptr->RequestHostName,
                          tkptr->ProxyLookupRetryCount,
                          &ProxyNetHostConnect,
                          tkptr);
      return;
   }

   /*********************************************************/
   /* this proxy server requests from another proxy server! */
   /*********************************************************/

   if (WATCHING(tkptr) && WATCH_CATEGORY(WATCH_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_PROXY,
                 "HOST-PROXY-CHAIN !AZ", tkptr->ChainHostPortPtr);

   tkptr->ConnectPort = tkptr->ChainIpPort;
   tkptr->ConnectHostPortPtr = tkptr->ChainHostPortPtr;

   tkptr->HostLookup.LookupIOsb.Status = SS$_NORMAL;
   ProxyNetHostConnect (tkptr);
}


/****************************************************************************/
/*
Called as an AST by TcpIpNameToAddress().  Check that the host name has been
resolved.  If not report the error. Create a socket and attempt to connect to
the remote, proxied server host.  AST to ProxyNetHostConnectAst().
*/

ProxyNetHostConnect (PROXY_TASK *tkptr)

{
   static BOOL  UseFullDuplexClose = true;

   int  status;
   unsigned long  *gwcptr;
   void  *BindSocketNamePtr;
   IO_SB  IOsb;
   REQUEST_STRUCT  *rqptr;
   SOCKADDRIN  *sin4ptr;
   SOCKADDRIN6  *sin6ptr;
   TCP_SOCKET_ITEM  *TcpSocketPtr;
   VMS_ITEM_LIST2  *il2ptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyNetHostConnect() !&Z !&S",
                 tkptr->ConnectHostPortPtr,
                 tkptr->HostLookup.LookupIOsb.Status);

   rqptr = tkptr->RequestPtr;

   if (VMSnok (tkptr->HostLookup.LookupIOsb.Status))
   {
      /*********************************/
      /* host address resolution error */
      /*********************************/

      if (WATCHING(tkptr) && WATCH_CATEGORY(WATCH_PROXY))
         WatchThis (WATCHTK(tkptr), FI_LI, WATCH_PROXY,
                    "HOST-LOOKUP !AZ !&S !-%!&M",
                    tkptr->RequestHostName,
                    tkptr->HostLookup.LookupIOsb.Status);

      /* dispose of the non-connected channel/socket used for the lookup */
      ProxyNetCloseSocket (tkptr);

      tkptr->ResponseStatusCode = 502;
      if (rqptr)
      {
         /* request associated with task */
         rqptr->rqResponse.HttpStatus = 502;
         rqptr->rqResponse.ErrorTextPtr = tkptr->RequestHostPort;
         if (tkptr->HostLookup.LookupIOsb.Status == SS$_ENDOFFILE)
            ErrorGeneral (rqptr,  MsgFor(rqptr,MSG_PROXY_HOST_UNKNOWN), FI_LI);
         else
            ErrorVmsStatus (rqptr, tkptr->HostLookup.LookupIOsb.Status, FI_LI);
      }
      ProxyEnd (tkptr);
      return;
   }

   /*****************/
   /* host resolved */
   /*****************/

   IPADDRESS_COPY (&tkptr->RequestHostIpAddress, &tkptr->HostLookup.IpAddress)

   /* if not chaining use resolved address, otherwise the specified chain */
   if (IPADDRESS_IS_RESET (&tkptr->ChainIpAddress))
      IPADDRESS_COPY (&tkptr->ConnectIpAddress, &tkptr->HostLookup.IpAddress)
   else
      IPADDRESS_COPY (&tkptr->ConnectIpAddress, &tkptr->ChainIpAddress)

   if (ProxyConnectPersistMax)
   {
      /* search for an existing connection */
      ProxyNetConnectSearch (tkptr);
      if (tkptr->ProxyChannel)
      {
         /***************************************/
         /* persistent proxy->origin connection */
         /***************************************/

         /* peek to ensure it's still connected and there's nothing buffered */
         static char  ThrowAwayBuffer [1];
         status = sys$qio (EfnNoWait, tkptr->ProxyChannel,
                           IO$_READVBLK, &tkptr->ProxyConnectIOsb,
                           &ProxyNetHostConnectAst, tkptr,
                           ThrowAwayBuffer, sizeof(ThrowAwayBuffer),
                           0, TCPIP$C_MSG_PEEK|TCPIP$C_MSG_NBIO, 0, 0);
         if (VMSnok (status))
         {
            /* leave it to the AST function to report! */
            tkptr->ProxyConnectIOsb.Status = status;
            SysDclAst (ProxyNetHostConnectAst, tkptr);
         }
         return;
      }
   }

   /*******************/
   /* connect to host */
   /*******************/

   if (WATCHING(tkptr) && WATCH_CATEGORY(WATCH_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_PROXY,
                 "CONNECT !&I,!UL as !&I", &tkptr->ConnectIpAddress,
                 tkptr->ConnectPort, &tkptr->BindIpAddress);

   tkptr->ConnectionCount = 1;

   /* assign a channel to the internet template device */
   status = sys$assign (&TcpIpDeviceDsc, &tkptr->ProxyChannel, 0, 0);
   if (VMSnok (status))
   {
      /* leave it to the AST function to report! */
      tkptr->ProxyConnectIOsb.Status = status;
      SysDclAst (ProxyNetHostConnectAst, tkptr);
      return;
   }

   if (IPADDRESS_IS_SET (&tkptr->BindIpAddress))
   {
      /* bind the proxy socket to a specific IP address */
      if (IPADDRESS_IS_V4 (&tkptr->BindIpAddress))
      {
         SOCKADDRESS_ZERO4 (&tkptr->ProxyBindSocketName)
         sin4ptr = &tkptr->ProxyBindSocketName.sa.v4;
         sin4ptr->SIN$W_FAMILY = TCPIP$C_AF_INET;
         sin4ptr->SIN$W_PORT = 0;
         IPADDRESS_SET4 (sin4ptr->SIN$L_ADDR, &tkptr->BindIpAddress)

         il2ptr = &tkptr->ProxyBindSocketNameItem;
         il2ptr->buf_len = sizeof(SOCKADDRIN);
         il2ptr->item = 0;
         il2ptr->buf_addr = sin4ptr;
      }
      else
      if (IPADDRESS_IS_V6 (&tkptr->BindIpAddress))
      {
         SOCKADDRESS_ZERO6 (&tkptr->ProxyBindSocketName)
         sin6ptr = &tkptr->ProxyBindSocketName.sa.v6;
         sin6ptr->SIN6$B_FAMILY = TCPIP$C_AF_INET6;
         sin6ptr->SIN6$W_PORT = 0;
         IPADDRESS_SET6 (sin6ptr->SIN6$R_ADDR_OVERLAY.SIN6$T_ADDR,
                         &tkptr->BindIpAddress)

         il2ptr = &tkptr->ProxyBindSocketNameItem;
         il2ptr->buf_len = sizeof(SOCKADDRIN);
         il2ptr->item = 0;
         il2ptr->buf_addr = sin6ptr;
      }
      else
         ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);

      BindSocketNamePtr = (void*)il2ptr;
   }
   else
      BindSocketNamePtr = 0;

   if (IPADDRESS_IS_V4 (&tkptr->ConnectIpAddress))
      TcpSocketPtr = &TcpIpSocket4;
   else
   if (IPADDRESS_IS_V6 (&tkptr->ConnectIpAddress))
      TcpSocketPtr = &TcpIpSocket6;
   else
      ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);

   /* make the channel a TCP, connection-oriented socket */
   for (;;)
   { 
      status = sys$qiow (EfnWait, tkptr->ProxyChannel, IO$_SETMODE,
                         &tkptr->ProxyConnectIOsb, 0, 0,
                         TcpSocketPtr, 0, BindSocketNamePtr,
                         0, UseFullDuplexClose ?
                               &TcpIpFullDuplexCloseOption : 0, 0);
      if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
         WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                       "sys$qiow() !&S !&S",
                       status, tkptr->ProxyConnectIOsb.Status);

      if (VMSok (status) && VMSok (tkptr->ProxyConnectIOsb.Status)) break;
      if (!UseFullDuplexClose) break;
      UseFullDuplexClose = false;

      /* Multinet 3.2 UCX driver barfs on FULL_DUPLEX_CLOSE, try without */
      if (VMSok (status) && VMSnok (tkptr->ProxyConnectIOsb.Status))
      {
         /* assign a new channel before retrying */
         sys$dassgn (tkptr->ProxyChannel);
         status = sys$assign (&TcpIpDeviceDsc, &tkptr->ProxyChannel, 0, 0);
         if (VMSnok (status))
         {
            /* leave it to the AST function to report! */
            tkptr->ProxyConnectIOsb.Status = status;
            SysDclAst (ProxyNetHostConnectAst, tkptr);
            return;
         }
      }
   }

   /* it's a $QIOW so the IO status block is valid */
   if (VMSok (status) && VMSnok (tkptr->ProxyConnectIOsb.Status))
      status = tkptr->ProxyConnectIOsb.Status;
   if (VMSnok (status))
   {
      /* leave it to the AST function to report! */
      tkptr->ProxyConnectIOsb.Status = status;
      SysDclAst (ProxyNetHostConnectAst, tkptr);
      return;
   }

   if (IPADDRESS_IS_V4 (&tkptr->ConnectIpAddress))
   {
      SOCKADDRESS_ZERO4 (&tkptr->ProxySocketName);
      sin4ptr = &tkptr->ProxySocketName.sa.v4;
      sin4ptr->SIN$W_FAMILY = TCPIP$C_AF_INET;
      sin4ptr->SIN$W_PORT = htons (tkptr->ConnectPort);
      IPADDRESS_SET4 (sin4ptr->SIN$L_ADDR, &tkptr->ConnectIpAddress)

      il2ptr = &tkptr->ProxySocketNameItem;
      il2ptr->buf_len = sizeof(SOCKADDRIN);
      il2ptr->item = TCPIP$C_SOCK_NAME;
      il2ptr->buf_addr = sin4ptr;
   }
   else
   if (IPADDRESS_IS_V6 (&tkptr->ConnectIpAddress))
   {
      SOCKADDRESS_ZERO6 (&tkptr->ProxySocketName);
      sin6ptr = &tkptr->ProxySocketName.sa.v6;
      sin6ptr->SIN6$B_FAMILY = TCPIP$C_AF_INET6;
      sin6ptr->SIN6$W_PORT = htons (tkptr->ConnectPort);
      IPADDRESS_SET6 (sin6ptr->SIN6$R_ADDR_OVERLAY.SIN6$T_ADDR,
                      &tkptr->ConnectIpAddress)

      il2ptr = &tkptr->ProxySocketNameItem;
      il2ptr->buf_len = sizeof(SOCKADDRIN6);
      il2ptr->item = TCPIP$C_SOCK_NAME;
      il2ptr->buf_addr = sin6ptr;
   }
   else
      ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);

   status = sys$qio (EfnNoWait, tkptr->ProxyChannel, IO$_ACCESS,
                     &tkptr->ProxyConnectIOsb, &ProxyNetHostConnectAst, tkptr,
                     0, 0, &tkptr->ProxySocketNameItem, 0, 0, 0);

   if (VMSnok (status))
   {
      /* leave it to the AST function to report! */
      tkptr->ProxyConnectIOsb.Status = status;
      SysDclAst (ProxyNetHostConnectAst, tkptr);
      return;
   }
}

/****************************************************************************/
/*
Called as an AST from ProxyNetHostConnect().  The remote server host connect
attempt has completed and either been successful or returned an error status. 
If an error then explain it and end proxy processing.  If successful then begin
processing the request scheme.
*/

ProxyNetHostConnectAst (PROXY_TASK *tkptr)

{
   int  status;
   char  *cptr;
   REQUEST_STRUCT  *rqptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyNetHostConnectAst() !&F !&Z !&S", ProxyNetHostConnectAst,
                 tkptr->ConnectHostPortPtr, tkptr->ProxyConnectIOsb.Status);

   rqptr = tkptr->RequestPtr;

   if (tkptr->ConnectionCount > 1)
   {
      /***************************************/
      /* persistent proxy->origin connection */
      /***************************************/

      /* of course shouldn't be able to read anything at all at this stage!! */
      if (VMSnok (tkptr->ProxyConnectIOsb.Status))
      {
         if (tkptr->ProxyConnectIOsb.Status == SS$_SUSPENDED)
         {
            /* this means the non-blocking peek found nothing buffered */
            tkptr->ProxyConnectIOsb.Status = SS$_NORMAL;
         }
         else
         {
            /* assume just a remote disconnection, close and restart */
            ProxyNetCloseSocket (tkptr);
            ProxyNetHostConnect (tkptr);
            return;
         }
      }
      else
      {
         /* so now that we did consider it an error, close and restart */
         ProxyNetCloseSocket (tkptr);
         ProxyNetHostConnect (tkptr);
         return;
      }
   }

   /*************************/
   /* IP version accounting */
   /*************************/

   /* after any connection persistence has been taken into account */
   if (IPADDRESS_IS_V4 (&tkptr->ConnectIpAddress))
   {
      InstanceMutexLock (INSTANCE_MUTEX_HTTPD);
      ProxyAccountingPtr->ConnectIpv4Count++;
      /* [ipv4=0,ipv6=1][ipv4=0,ipv6=1] */
      if (IPADDRESS_IS_V4 (&tkptr->ServicePtr->ServerIpAddress))
         ProxyAccountingPtr->GatewayIpvCount[0][0]++;
      else
         ProxyAccountingPtr->GatewayIpvCount[1][0]++;
      InstanceMutexUnLock (INSTANCE_MUTEX_HTTPD);
   }
   else
   if (IPADDRESS_IS_V6 (&tkptr->ConnectIpAddress))
   {
      InstanceMutexLock (INSTANCE_MUTEX_HTTPD);
      ProxyAccountingPtr->ConnectIpv6Count++;
      /* [ipv4=0,ipv6=1][ipv4=0,ipv6=1] */
      if (IPADDRESS_IS_V4 (&tkptr->ServicePtr->ServerIpAddress))
         ProxyAccountingPtr->GatewayIpvCount[0][1]++;
      else
         ProxyAccountingPtr->GatewayIpvCount[1][1]++;
      InstanceMutexUnLock (INSTANCE_MUTEX_HTTPD);
   }
   else
      ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);

   if (VMSnok (tkptr->ProxyConnectIOsb.Status))
   {
      /*****************/
      /* connect error */
      /*****************/

      if (WATCHING(tkptr) && WATCH_CATEGORY(WATCH_PROXY))
         WatchThis (rqptr, FI_LI, WATCH_PROXY,
                    "CONNECT !&S %!&M",
                    tkptr->ProxyConnectIOsb.Status,
                    tkptr->ProxyConnectIOsb.Status);

      /* dispose of the non-connected channel/socket used for the lookup */
      ProxyNetCloseSocket (tkptr);

      tkptr->ResponseStatusCode = 502;
      if (rqptr)
      {
         /* request associated with task */
         rqptr->rqResponse.HttpStatus = 502;

         switch (tkptr->ProxyConnectIOsb.Status)
         {
            case PROXY_ERROR_CONNECT_REFUSED :

               if (IPADDRESS_IS_SET (&tkptr->ChainIpAddress))
                  cptr = MsgFor(rqptr,MSG_PROXY_CHAIN_REFUSED);
               else
                  cptr = MsgFor(rqptr,MSG_PROXY_CONNECT_REFUSED);

               ErrorGeneral (rqptr, cptr, FI_LI);
               break;

            case PROXY_ERROR_HOST_UNREACHABLE :

               if (IPADDRESS_IS_SET (&tkptr->ChainIpAddress))
                  cptr = MsgFor(rqptr,MSG_PROXY_CHAIN_UNREACHABLE);
               else
                  cptr = MsgFor(rqptr,MSG_PROXY_HOST_UNREACHABLE);

               ErrorGeneral (rqptr, cptr, FI_LI);
               break;

            case PROXY_ERROR_HOST_TIMEOUT :

               if (IPADDRESS_IS_SET (&tkptr->ChainIpAddress))
                  cptr = MsgFor(rqptr,MSG_PROXY_CHAIN_UNREACHABLE);
               else
                  cptr = MsgFor(rqptr,MSG_PROXY_HOST_UNREACHABLE);

               tkptr->ResponseStatusCode = 504;
               ErrorGeneral (rqptr, cptr, FI_LI);
               break;

            default :

               if (IPADDRESS_IS_SET (&tkptr->ChainIpAddress))
               {
                  rqptr->rqResponse.ErrorTextPtr =
                     MsgFor(rqptr,MSG_PROXY_CHAIN_FAILURE);
                  rqptr->rqResponse.ErrorOtherTextPtr =
                     tkptr->ChainHostPortPtr;
               }
               else
                  rqptr->rqResponse.ErrorTextPtr = tkptr->RequestHostPort;
               ErrorVmsStatus (rqptr, tkptr->ProxyConnectIOsb.Status, FI_LI);
         }
      }

      /* invalidate any entry (still) in the host name/address cache */
      TcpIpCacheSetEntry (tkptr->RequestHostName,
                          strlen(tkptr->RequestHostName),
                          NULL);

      ProxyEnd (tkptr);
      return;
   }

   if (tkptr->ProxyTunnel)
   {
      /**************/
      /* tunnelling */
      /**************/

      if (WATCHING(tkptr) && WATCH_CATEGORY(WATCH_PROXY))
      {
         switch (tkptr->ProxyTunnel)
         {
            case PROXY_TUNNEL_CONNECT : cptr = "CONNECT"; break;
            case PROXY_TUNNEL_HTTP : cptr = "HTTP"; break;
            case PROXY_TUNNEL_HTTPS : cptr = "HTTPS"; break;
            case PROXY_TUNNEL_RAW : cptr = "RAW"; break;
            default : cptr = "?";
         }
         WatchThis (WATCHTK(tkptr), FI_LI, WATCH_PROXY,
                    "TUNNEL (!AZ) established", cptr);
      }

      /* it's successful if the remote host accepts it */
      if (rqptr) rqptr->rqResponse.HttpStatus = 200;

      /* once we start reading into the buffer the request header is kaput */
      rqptr->rqHeader.RequestHeaderPtrInvalid = true;

      if (tkptr->ProxyTunnel == PROXY_TUNNEL_HTTPS)
      {
         if (tkptr->ServicePtr->SSLclientEnabled)
            SesolaNetClientBegin (tkptr);
         else
         {
            if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_PROXY))
               WatchThis (rqptr, FI_LI, WATCH_PROXY,
                          "CLIENT SSL not configured");
            rqptr->rqResponse.HttpStatus = 503;
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_PROXY_GATEWAY), FI_LI);
            ProxyEnd (rqptr);
            return;
         }
      }
      else
         ProxyTunnelBegin (tkptr);
   }
   else
   {
      /*************************************/ 
      /* HTTP "GET", "POST", etc., methods */
      /*************************************/ 

      if (tkptr->ConnectPort == DEFAULT_HTTPS_PORT ||
          tkptr->RequestScheme == PROXY_SCHEME_HTTPSSL)
      {
         if (tkptr->ServicePtr->SSLclientEnabled)
            SesolaNetClientBegin (tkptr);
         else
         {
            if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_PROXY))
               WatchThis (rqptr, FI_LI, WATCH_PROXY,
                          "CLIENT SSL not configured");
            rqptr->rqResponse.HttpStatus = 503;
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_PROXY_GATEWAY), FI_LI);
            ProxyEnd (rqptr);
            return;
         }
      }
      else
      if (tkptr->RequestScheme == PROXY_SCHEME_HTTP)
         ProxyWriteRequest (tkptr);
      else
      if (tkptr->RequestScheme == PROXY_SCHEME_FTP)
         ProxyFtpLifeCycle (tkptr);
      else
         ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
   }
}

/*****************************************************************************/
/*
Write data to the proxied server. Explicitly declares any AST routine if an
error occurs. The calling function must not do any error recovery if an AST
routine has been supplied but the associated AST routine must! If an AST was
not supplied then the return status can be checked.  AST calls
ProxyNetWriteRawAST() which will then call the supplied AST function.
*/

int ProxyNetWriteRaw
(
PROXY_TASK *tkptr,
PROXY_AST AstFunction,
char *DataPtr,
int DataLength
)

{
   int  status;
   REQUEST_STRUCT *rqptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyNetWriteRaw() !&F !&A !&X !UL",
                 &ProxyNetWriteRaw, AstFunction, DataPtr, DataLength);

   rqptr = tkptr->RequestPtr;

   if (tkptr->ProxyWriteRawAstFunction ||
       !DataLength)
   {
      tkptr->ProxyWriteIOsb.Status = SS$_BUGCHECK;
      tkptr->ProxyWriteIOsb.Count = 0;
      if (AstFunction) SysDclAst (AstFunction, tkptr);
      return (tkptr->ProxyWriteIOsb.Status);
   }
   tkptr->ProxyWriteRawAstFunction = AstFunction;
   tkptr->ProxyWriteRawDataPtr = DataPtr;
   tkptr->ProxyWriteRawDataCount = DataLength;

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_NETWORK_OCTETS))
   {
      WatchThis (rqptr, FI_LI, WATCH_NETWORK,
                 "WRITE !UL bytes", DataLength);
      WatchDataDump (DataPtr, DataLength);
   }

   if (!AstFunction)
   {
      /***************/
      /* blocking IO */
      /***************/

      status = sys$qiow (EfnWait, tkptr->ProxyChannel,
                         IO$_WRITEVBLK, &tkptr->ProxyWriteIOsb, 0, 0,
                         DataPtr, DataLength, 0, 0, 0, 0);
   }
   else
   {
      /*******************/
      /* non-blocking IO */
      /*******************/

      status = sys$qio (EfnNoWait, tkptr->ProxyChannel,
                        IO$_WRITEVBLK, &tkptr->ProxyWriteIOsb,
                        &ProxyNetWriteRawAst, tkptr,
                        DataPtr, DataLength, 0, 0, 0, 0);
   }

   /****************/
   /* check status */
   /****************/

   if (VMSok (status)) return (status);

   /* if resource wait enabled the only quota not waited for is ASTLM */
   if (status == SS$_EXQUOTA)
      ErrorExitVmsStatus (status, "sys$qio()", FI_LI);

   /* write failed, call AST explicitly, status in the IOsb */
   tkptr->ProxyWriteIOsb.Status = status;
   tkptr->ProxyWriteIOsb.Count = 0;
   SysDclAst (ProxyNetWriteRawAst, tkptr);
   return (status);
}

/*****************************************************************************/
/*
AST from ProxyNetWriteRaw().  Call the AST function.
*/ 

ProxyNetWriteRawAst (PROXY_TASK *tkptr)

{
   int  status;
   PROXY_AST  AstFunction;
   REQUEST_STRUCT *rqptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyNetWriteRawAst() !&F !&S !UL", &ProxyNetWriteRawAst,
                 tkptr->ProxyWriteIOsb.Status, tkptr->ProxyWriteIOsb.Count);

   rqptr = tkptr->RequestPtr;

   if (WATCHING(tkptr) && WATCH_CATEGORY(WATCH_NETWORK))
   {
      if (VMSnok(tkptr->ProxyWriteIOsb.Status))
         WatchThis (WATCHTK(tkptr), FI_LI, WATCH_NETWORK,
                    "WRITE !&S %!&M",
                    tkptr->ProxyWriteIOsb.Status,
                    tkptr->ProxyWriteIOsb.Status);
      else
         WatchThis (WATCHTK(tkptr), FI_LI, WATCH_NETWORK,
                    "WRITE !&S !UL bytes",
                    tkptr->ProxyWriteIOsb.Status,
                    tkptr->ProxyWriteIOsb.Count);
   }

   if (VMSok (tkptr->ProxyWriteIOsb.Status))
   {
      ADD_LONG_QUAD (tkptr->ProxyWriteIOsb.Count, tkptr->BytesRawTx);
   }
   else
   {
      if (tkptr->ProxyWriteIOsb.Status != SS$_LINKDISCON &&
          tkptr->ProxyWriteIOsb.Status != SS$_CONNECFAIL &&
          tkptr->ProxyWriteIOsb.Status != SS$_ABORT &&
          tkptr->ProxyWriteIOsb.Status != SS$_CANCEL &&
          tkptr->ProxyWriteIOsb.Status != SS$_TIMEOUT &&
          tkptr->ProxyWriteIOsb.Status != SS$_IVCHAN)
         ErrorNoticed (rqptr, tkptr->ProxyWriteIOsb.Status,
                       "ProxyWriteIOsb", FI_LI);
   }

   tkptr->ProxyWriteRawDataPtr = tkptr->ProxyWriteRawDataCount = 0;
   if (!tkptr->ProxyWriteRawAstFunction) return;
   AstFunction = tkptr->ProxyWriteRawAstFunction;
   tkptr->ProxyWriteRawAstFunction = NULL;
   (*AstFunction)(tkptr);
}

/*****************************************************************************/
/*
Queue up a read from the proxied server over the network. If 'AstFunction' 
is zero then no I/O completion AST routine is called.  If it is non-zero then 
the function pointed to by the parameter is called when the network write 
completes.

Explicitly declares any AST routine if an error occurs. The calling function
must not do any error recovery if an AST routine has been supplied but the
associated AST routine must!  If an AST was not supplied then the return
status can be checked.
*/ 

int ProxyNetReadRaw
(
PROXY_TASK *tkptr,
PROXY_AST AstFunction,
char *DataPtr,
int DataSize
)
{
   int  status;
   REQUEST_STRUCT *rqptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyNetReadRaw() !&F !&A !&X !UL",
                 &ProxyNetReadRaw, AstFunction, DataPtr, DataSize);

   rqptr = tkptr->RequestPtr;

   if (tkptr->ProxyReadRawAstFunction)
   {
      tkptr->ProxyReadIOsb.Status = SS$_BUGCHECK;
      tkptr->ProxyReadIOsb.Count = 0;
      if (AstFunction) SysDclAst (AstFunction, tkptr);
      return (tkptr->ProxyReadIOsb.Status);
   }
   tkptr->ProxyReadRawAstFunction = AstFunction;
   tkptr->ProxyReadRawDataPtr = DataPtr;
   tkptr->ProxyReadRawDataSize = DataSize;

   if (WATCHING(rqptr) && WATCH_CATEGORY(WATCH_NETWORK_OCTETS))
      WatchThis (rqptr, FI_LI, WATCH_NETWORK,
                 "READ !UL bytes max", DataSize);

   if (!AstFunction)
   {
      /***************/
      /* blocking IO */
      /***************/

      status = sys$qiow (EfnWait, tkptr->ProxyChannel,
                         IO$_READVBLK, &tkptr->ProxyReadIOsb, 0, 0,
                         DataPtr, DataSize, 0, 0, 0, 0);
   }
   else
   {
      /*******************/
      /* non-blocking IO */
      /*******************/

      status = sys$qio (EfnNoWait, tkptr->ProxyChannel,
                        IO$_READVBLK, &tkptr->ProxyReadIOsb,
                        &ProxyNetReadRawAst, tkptr,
                        DataPtr, DataSize, 0, 0, 0, 0);
   }

   /****************/
   /* check status */
   /****************/

   if (VMSok (status)) return (status);

   /* with resource wait enabled the only quota not waited for is ASTLM */
   if (status == SS$_EXQUOTA)
      ErrorExitVmsStatus (status, "sys$qio()", FI_LI);

   /* queuing of read failed, call AST explicitly, status in the IOsb */
   tkptr->ProxyReadIOsb.Status = status;
   tkptr->ProxyReadIOsb.Count = 0;
   SysDclAst (ProxyNetReadRawAst, tkptr);
   return (status);
}

/*****************************************************************************/
/*
AST from ProxyNetReadRaw().  Call the AST function.
*/ 

ProxyNetReadRawAst (PROXY_TASK *tkptr)

{
   int  status;
   PROXY_AST  AstFunction;
   REQUEST_STRUCT *rqptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyNetReadRawAst() !&F !&S !UL", &ProxyNetReadRawAst,
                 tkptr->ProxyReadIOsb.Status, tkptr->ProxyReadIOsb.Count);

   rqptr = tkptr->RequestPtr;

   if (WATCHING(tkptr))
   {
      if (WATCH_CATEGORY(WATCH_NETWORK) ||
          WATCH_CATEGORY(WATCH_NETWORK_OCTETS))
      {
         if (VMSok(tkptr->ProxyReadIOsb.Status))
         {
            WatchThis (WATCHTK(tkptr), FI_LI, WATCH_NETWORK,
                       "READ !&S !UL bytes",
                       tkptr->ProxyReadIOsb.Status,
                       tkptr->ProxyReadIOsb.Count);

            if WATCH_CATEGORY(WATCH_NETWORK_OCTETS)
               WatchDataDump (tkptr->ProxyReadRawDataPtr,
                              tkptr->ProxyReadIOsb.Count);
         }
         else
            WatchThis (WATCHTK(tkptr), FI_LI, WATCH_NETWORK,
                       "READ !&S %!&M",
                       tkptr->ProxyReadIOsb.Status,
                       tkptr->ProxyReadIOsb.Status);
      }
   }

   if (VMSok (tkptr->ProxyReadIOsb.Status))
   {
      ADD_LONG_QUAD (tkptr->ProxyReadIOsb.Count, tkptr->BytesRawRx);
      /* zero bytes with a normal status is a definite no-no (TGV-Multinet) */
      if (!tkptr->ProxyReadIOsb.Count) tkptr->ProxyReadIOsb.Status = SS$_ABORT;
   }
   else
   {
      if (tkptr->ProxyReadIOsb.Status != SS$_LINKDISCON &&
          tkptr->ProxyReadIOsb.Status != SS$_CONNECFAIL &&
          tkptr->ProxyReadIOsb.Status != SS$_ABORT &&
          tkptr->ProxyReadIOsb.Status != SS$_CANCEL &&
          tkptr->ProxyReadIOsb.Status != SS$_TIMEOUT &&
          tkptr->ProxyReadIOsb.Status != SS$_IVCHAN)
         ErrorNoticed (rqptr, tkptr->ProxyReadIOsb.Status,
                       "ProxyReadIOsb", FI_LI);
   }

   tkptr->ProxyReadRawDataPtr = tkptr->ProxyReadRawDataSize = 0;
   if (!tkptr->ProxyReadRawAstFunction) return;
   AstFunction = tkptr->ProxyReadRawAstFunction;
   tkptr->ProxyReadRawAstFunction = NULL;
   (*AstFunction)(tkptr);
}

/****************************************************************************/
/*
Just shut the socket down, bang!
*/

int ProxyNetCloseSocket (PROXY_TASK *tkptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyNetCloseSocket() !&F", &ProxyNetCloseSocket);

   if (!tkptr->ProxyChannel) return (SS$_NORMAL);

   status = sys$dassgn (tkptr->ProxyChannel);
   tkptr->ProxyChannel = 0;
   if (tkptr->SesolaPtr) SesolaNetSocketHasBeenClosed (tkptr->SesolaPtr);

   if (WATCHING(tkptr) && WATCH_CATEGORY(WATCH_PROXY))
   {
      if (VMSok(status))
         WatchThis (WATCHTK(tkptr), FI_LI, WATCH_PROXY,
                    "CLOSE !AZ,!UL",
                    tkptr->RequestHostName, tkptr->RequestPort); 
      else
         WatchThis (WATCHTK(tkptr), FI_LI, WATCH_PROXY,
                    "CLOSE !AZ,!UL !&S %!&M",
                    tkptr->RequestHostName, tkptr->RequestPort,
                    status, status);
   }

   return (status);
}

/****************************************************************************/
/*
Create a persistent connection structure (using process memory) and add it to
the list.  Copy the channel number from the proxy task socket into the
structure (zeroing the task channel), along with the IP address and port of the
connection.
*/

ProxyNetConnectPersist (PROXY_TASK *tkptr)

{
   int  status;
   PROXY_CONNECT  *pcptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyNetConnectPersist() !&I,!UL !UL",
                 &tkptr->ConnectIpAddress, tkptr->ConnectPort,
                 tkptr->ConnectionCount);

   if (ProxyNetConnectCount + ProxyNetConnectFreeCount >=
       ProxyConnectPersistMax)
   {
      InstanceGblSecIncrLong (&ProxyAccountingPtr->ConnectPersistFull);
      return;
   }

   if (pcptr = ProxyNetConnectFreeList.HeadPtr)
   {
      ListRemove (&ProxyNetConnectFreeList, pcptr);
      memset (pcptr, 0, sizeof(PROXY_CONNECT));
      if (ProxyNetConnectFreeCount) ProxyNetConnectFreeCount--;
   }
   else
      pcptr = (PROXY_CONNECT*)VmGet(sizeof(PROXY_CONNECT));

   pcptr->EntryTickSecond = HttpdTickSecond;
   pcptr->ProxyChannel = tkptr->ProxyChannel;
   IPADDRESS_COPY (&pcptr->ConnectIpAddress, &tkptr->ConnectIpAddress);
   pcptr->ConnectPort = tkptr->ConnectPort;
   pcptr->ConnectionCount = tkptr->ConnectionCount;

   ListAddTail (&ProxyNetConnectList, pcptr);

   ProxyNetConnectCount++;
   InstanceMutexLock (INSTANCE_MUTEX_HTTPD);
   if (ProxyNetConnectCount > ProxyAccountingPtr->ConnectPersistPeak)
      ProxyAccountingPtr->ConnectPersistPeak = ProxyNetConnectCount;
   InstanceMutexUnLock (INSTANCE_MUTEX_HTTPD);

   /* the task channel is now declared unused */
   tkptr->ProxyChannel = 0;
}

/****************************************************************************/
/*
Search through any list of persistent connections comparing the requests IP
address and port to that stored in each list entry.  If one matched copy the
channel and persistence count to the proxy task and remove the entry from the
list.  Returns with the task connection socket channel non-zero if it found a
match or still as zero if not.
*/

ProxyNetConnectSearch (PROXY_TASK *tkptr)

{
   int  status;
   LIST_ENTRY  *leptr;
   PROXY_CONNECT  *pcptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY,
                 "ProxyNetConnectSearch() !&I,!UL !UL",
                 &tkptr->ConnectIpAddress, tkptr->ConnectPort,
                 ProxyNetConnectCount);

   /* important these be zeroed to indicate search failure */
   tkptr->ProxyChannel = tkptr->ConnectionCount = 0;

   if (!ProxyNetConnectCount) return;

   for (leptr = ProxyNetConnectList.HeadPtr; leptr; leptr = leptr->NextPtr)
   {
      pcptr = (PROXY_CONNECT*)leptr;

      if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
         WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY, "!&I,!UL !UL !UL",
                    &pcptr->ConnectIpAddress, pcptr->ConnectPort,
                    pcptr->ConnectionCount,
                    pcptr->ProxyChannel);

      if (!IPADDRESS_IS_SAME (&pcptr->ConnectIpAddress,
                              &tkptr->ConnectIpAddress)) continue;
      if (pcptr->ConnectPort != tkptr->ConnectPort) continue;

      tkptr->ProxyChannel = pcptr->ProxyChannel;
      tkptr->ConnectionCount = pcptr->ConnectionCount + 1;

      ListRemove (&ProxyNetConnectList, pcptr);
      if (ProxyNetConnectCount) ProxyNetConnectCount--;
      ListAddTail (&ProxyNetConnectFreeList, pcptr);
      ProxyNetConnectFreeCount++;

      InstanceMutexLock (INSTANCE_MUTEX_HTTPD);
      ProxyAccountingPtr->ConnectPersistCount++;
      if (tkptr->ConnectionCount >
          ProxyAccountingPtr->ConnectPersistMax)
         ProxyAccountingPtr->ConnectPersistMax =
            tkptr->ConnectionCount;
      InstanceMutexUnLock (INSTANCE_MUTEX_HTTPD);

      if (WATCHING(tkptr) && WATCH_CATEGORY(WATCH_PROXY))
         WatchThis (WATCHTK(tkptr), FI_LI, WATCH_PROXY,
                    "PERSISTENT !UL !AZ,!UL",
                    tkptr->ConnectionCount,
                    tkptr->RequestHostName, tkptr->RequestPort); 
      return;
   }

   if (WATCHING(tkptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (WATCHTK(tkptr), FI_LI, WATCH_MOD_PROXY, "NOT FOUND");
}

/****************************************************************************/
/*
Called by HttpdTick() once a second.  Scan the list of persistent proxy->origin
connections looking for those that have timed-out.  For each timeout just close
the socket by deassigning the channel and remove and free the entry from the
list.  If 'TimeoutSeconds' is negative use the configuration parameter.  If
zero then timeout all of the connections (purge the list).  If a positive
integer use this as the timeout period.
*/

int ProxyNetConnectSupervisor (int TimeoutPeriod)

{
   int  status,
        TimeoutSeconds;
   LIST_ENTRY  *leptr;
   PROXY_CONNECT  *pcptr;

   /*********/
   /* begin */
   /*********/

/**
   if (WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (NULL, FI_LI, WATCH_MOD_PROXY,
                 "ProxyNetConnectSupervisor() !SL !UL !UL !UL",
                 TimeoutPeriod, ProxyConnectPersistSeconds,
                 HttpdTickSecond, ProxyNetConnectCount);
**/

   if (!ProxyNetConnectCount) return (false);

   if (!TimeoutPeriod)
      TimeoutSeconds = 0;
   else
   if (TimeoutPeriod < 0)
      TimeoutSeconds = ProxyConnectPersistSeconds;
   else
      TimeoutSeconds = TimeoutPeriod;

   for (leptr = ProxyNetConnectList.HeadPtr; leptr; leptr = leptr->NextPtr)
   {
      pcptr = (PROXY_CONNECT*)leptr;

      if (WATCH_MODULE(WATCH_MOD_PROXY))
         WatchThis (NULL, FI_LI, WATCH_MOD_PROXY,
                    "!&I,!UL !UL+!UL(!&B) !UL !UL",
                    &pcptr->ConnectIpAddress, pcptr->ConnectPort,
                    pcptr->EntryTickSecond, TimeoutSeconds,
                    pcptr->EntryTickSecond + TimeoutSeconds <= HttpdTickSecond,
                    pcptr->ConnectionCount,
                    pcptr->ProxyChannel);

      if (pcptr->EntryTickSecond + TimeoutSeconds < HttpdTickSecond)
      {
         status = sys$dassgn (pcptr->ProxyChannel);
         if (VMSnok (status))
            ErrorNoticed (NULL, status, "sys$dassgn()", FI_LI);
         ListRemove (&ProxyNetConnectList, pcptr);
         if (ProxyNetConnectCount) ProxyNetConnectCount--;
         ListAddTail (&ProxyNetConnectFreeList, pcptr);
         ProxyNetConnectFreeCount++;
      }
   }

   if (ProxyNetConnectCount) return (true);

   return (false);
}

/*****************************************************************************/
/*
Return a report on proxy->origin server persistent connections.
*/ 

ProxyNetPersistentReport
(
REQUEST_STRUCT *rqptr,
REQUEST_AST NextTaskFunction
)
{
   static char  BeginPageFao [] =
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Persist:</TH><TD ALIGN=left COLSPAN=2>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Limit:</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Current:</TH>\
<TD ALIGN=right>!UL</TD><TD ALIGN=left>(!UL)</TD></TR>\n\
<TR><TH ALIGN=right>Full:</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Peak:</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Max:</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Count:</TH>\
<TD ALIGN=right>!UL</TD><TD ALIGN=left>(!UL%)</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=1 CELLSPACING=0 BORDER=0>\n\
<TR>\
<TH></TH>\
<TH ALIGN=left><U>Address</U>&nbsp;&nbsp;</TH>\
<TH ALIGN=right><U>Port</U>&nbsp;&nbsp;</TH>\
<TH ALIGN=right><U>Count</U>&nbsp;&nbsp;</TH>\
<TH ALIGN=right><U>Persist</U></TH>\
<TH>&nbsp;</TH>\
</TR>\n\
<TR HEIGHT=5></TR>\n";

   static char  ItemFao [] =
"<TR><TH ALIGN=right>!3ZL&nbsp;&nbsp;</TH>\
<TD ALIGN=left>!&I&nbsp;&nbsp;</TD>\
<TD ALIGN=right>!UL&nbsp;&nbsp;</TD>\
<TD ALIGN=right>!UL&nbsp;&nbsp;</TD>\
<TD ALIGN=right>!AZ</TD>\
<TR>\n";

   static char  ItemOddFao [] =
"<TR><TH ALIGN=right>!3ZL&nbsp;&nbsp;</TH>\
<TD BGCOLOR=\"#eeeeee\" ALIGN=left>!&I&nbsp;&nbsp;</TD>\
<TD BGCOLOR=\"#eeeeee\" ALIGN=right>!UL&nbsp;&nbsp;</TD>\
<TD BGCOLOR=\"#eeeeee\" ALIGN=right>!UL&nbsp;&nbsp;</TD>\
<TD BGCOLOR=\"#eeeeee\" ALIGN=right>!AZ</TD>\
<TR>\n";

   static char EmptyListFao [] =
"<TR><TH>000&nbsp;&nbsp;</TH>\
<TD COLSPAN=4 BGCOLOR=\"#eeeeee\"><I>empty</I></TD><TR>\n";

   static char  EndPageFao [] =
"</TABLE>\n\
<HR SIZE=1 NOSHADE WIDTH=60% ALIGN=left>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TD>\n\
\
</TD></TR>\n\
</TABLE>\n\
</BODY>\n\
</HTML>\n";

   int  status,
        CountItem,
        CountNetwork,
        CountPercent;
   unsigned short  Length;
   unsigned long  FaoVector [32];
   unsigned long  *vecptr;
   LIST_ENTRY  *leptr;
   PROXY_CONNECT  *pcptr;

   /*********/
   /* begin */
   /*********/

   if (WATCHING(rqptr) && WATCH_MODULE(WATCH_MOD_PROXY))
      WatchThis (rqptr, FI_LI, WATCH_MOD_PROXY,
                 "ProxyNetPersistentReport() !&A", NextTaskFunction);

   AdminPageTitle (rqptr, "Proxy Connection Report");


   InstanceMutexLock (INSTANCE_MUTEX_HTTPD);

   CountNetwork = ProxyAccountingPtr->ConnectIpv4Count +
                  ProxyAccountingPtr->ConnectIpv6Count;

   if (CountNetwork)
      CountPercent =
         (int)(((float)ProxyAccountingPtr->ConnectPersistCount * 100) /
               (float)CountNetwork);
   else
      CountPercent = 0;

   vecptr = FaoVector;

   *vecptr++ = MetaConShowSeconds (rqptr, ProxyConnectPersistSeconds);
   *vecptr++ = ProxyConnectPersistMax;
   *vecptr++ = ProxyNetConnectCount;
   *vecptr++ = ProxyNetConnectFreeCount;
   *vecptr++ = ProxyAccountingPtr->ConnectPersistFull;
   *vecptr++ = ProxyAccountingPtr->ConnectPersistPeak;
   *vecptr++ = ProxyAccountingPtr->ConnectPersistMax;
   *vecptr++ = ProxyAccountingPtr->ConnectPersistCount;
   *vecptr++ = CountPercent;

   InstanceMutexUnLock (INSTANCE_MUTEX_HTTPD);

   status = NetWriteFaol (rqptr, BeginPageFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (rqptr, status, "NetWriteFaol()", FI_LI);

   CountItem = 0;
   for (leptr = ProxyNetConnectList.HeadPtr; leptr; leptr = leptr->NextPtr)
   {
      pcptr = (PROXY_CONNECT*)leptr;

      vecptr = FaoVector;

      *vecptr++ = ++CountItem;
      *vecptr++ = &pcptr->ConnectIpAddress;
      *vecptr++ = pcptr->ConnectPort;
      *vecptr++ = pcptr->ConnectionCount;
      *vecptr++ = MetaConShowSeconds (rqptr,
                     HttpdTickSecond - pcptr->EntryTickSecond);

      if (CountItem & 1)
         status = NetWriteFaol (rqptr, ItemOddFao, &FaoVector);
      else
         status = NetWriteFaol (rqptr, ItemFao, &FaoVector);
      if (VMSnok (status)) ErrorNoticed (rqptr, status, "NetWriteFaol()", FI_LI);
   }
   if (!ProxyNetConnectList.HeadPtr)
   {
      status = NetWriteFaol (rqptr, EmptyListFao, NULL);
      if (VMSnok (status)) ErrorNoticed (rqptr, status, "NetWriteFaol()", FI_LI);
   }

   status = NetWriteFaol (rqptr, EndPageFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (rqptr, status, "NetWriteFaol()", FI_LI);

   rqptr->rqResponse.PreExpired = PRE_EXPIRE_ADMIN;
   RESPONSE_HEADER_200_HTML_BUFFER (rqptr);

   SysDclAst (NextTaskFunction, rqptr);
}

/****************************************************************************/

